#!/bin/bash

set -xeuo pipefail
IFS=$'\n\t'

DIR="$(realpath ${0%/*})"
RUNTIME_IS_PODMAN="false"
if [[ -x "$(which podman)" ]]; then
  CONTAINER_RUNTIME="$(which podman 2> /dev/null)"
  RUNTIME_IS_PODMAN="true"
elif [[ -x "$(which docker)" ]]; then
  CONTAINER_RUNTIME="$(which docker 2> /dev/null)"
else
  echo "Container runtime (docker/podman) not found!"
  exit 1
fi

# replace containers
if $CONTAINER_RUNTIME inspect --type container synapse 1> /dev/null; then
  $CONTAINER_RUNTIME rm -f synapse
fi
if $CONTAINER_RUNTIME inspect --type container nextcloud 1> /dev/null; then
  $CONTAINER_RUNTIME rm -f nextcloud
fi

if [[ ! -d vendor ]]; then
    make composer
fi

$CONTAINER_RUNTIME run -d \
    --name=nextcloud \
    -p 8080:80 \
    -p 8008:8008 \
    -v "$DIR:/var/www/html/custom_apps/upschooling" \
    --add-host "synapse:127.0.0.1" \
    docker.io/nextcloud

$CONTAINER_RUNTIME exec nextcloud bash -c 'apt-get update && apt-get install -y git'
$CONTAINER_RUNTIME exec nextcloud chown -R 33 /var/www/html/custom_apps

if [[ $(id -u) -ne 0 && $RUNTIME_IS_PODMAN == "true" ]]; then
    $CONTAINER_RUNTIME unshare -- chown -R 33 "$DIR"
    $CONTAINER_RUNTIME unshare -- chgrp -R 0 "$DIR"
    $CONTAINER_RUNTIME unshare -- chmod -R ug+rw "$DIR"
fi

$CONTAINER_RUNTIME exec --user 33 nextcloud bash -c 'cd /var/www/html/custom_apps/upschooling && make composer'
$CONTAINER_RUNTIME exec --user 33 nextcloud php occ maintenance:install --database sqlite --admin-user admin --admin-pass admin
$CONTAINER_RUNTIME exec --user 33 nextcloud php occ config:system:set --value=true --type=boolean debug
$CONTAINER_RUNTIME exec --user 33 nextcloud php occ app:enable --force upschooling

SYNAPSE_DATA_VOLUME_EXISTS="false"
if [[ $RUNTIME_IS_PODMAN == "true" ]]; then
  if $CONTAINER_RUNTIME volume exists synapse-data; then
    echo "Found existing synapse-data volume (via volume exists)"
    SYNAPSE_DATA_VOLUME_EXISTS="true"
  fi
else
  if $CONTAINER_RUNTIME volume inspect synapse-data 1> /dev/null; then
    echo "Found existing synapse-data volume (via volume inspect)"
    SYNAPSE_DATA_VOLUME_EXISTS="true"
  fi
fi

if [[ $SYNAPSE_DATA_VOLUME_EXISTS == "false" ]]; then
  $CONTAINER_RUNTIME run --rm \
      --name=synapse \
      --hostname synapse \
      "--mount=type=volume,src=synapse-data,dst=/data" \
      -e SYNAPSE_SERVER_NAME=synapse \
      -e SYNAPSE_REPORT_STATS=no \
      docker.io/matrixdotorg/synapse \
      generate
  echo "Generated fresh synapse-data volume"
fi

LOCAL_SYNAPSE_DATA_PATH="$($CONTAINER_RUNTIME volume inspect --format "{{.Mountpoint}}" synapse-data)"
SED_ARGS=(-i 's|registration_shared_secret: .*|registration_shared_secret: "oyYh_iEJ7Aim.iB+ye.Xk;Gl3iHFab5*8K,zv~IulT85P=c-38"|' "$LOCAL_SYNAPSE_DATA_PATH/homeserver.yaml")
if [[ $(id -u) -eq 0 || $RUNTIME_IS_PODMAN != "true" ]]; then
    sed ${SED_ARGS[@]}
else
    podman unshare -- sed ${SED_ARGS[@]}
fi

$CONTAINER_RUNTIME run -d \
    --log-driver "k8s-file" \
    --name=synapse \
    "--mount=type=volume,src=synapse-data,dst=/data" \
    "--network=container:$($CONTAINER_RUNTIME inspect --format "{{.Id}}" nextcloud)" \
    --hostname synapse \
    docker.io/matrixdotorg/synapse

# wait for synapse to start
MAX_TRIES=15
for ((i = 0 ; i < $MAX_TRIES ; i++)); do
  if $CONTAINER_RUNTIME logs synapse 2>&1 | grep -q "Synapse now listening on TCP port 8008"; then
    echo -e "Synapse has started. \e[1;38;5;2mOK\033[0m"
    break
  fi
  sleep 1
done

if [[ $i -ge $MAX_TRIES ]]; then
  echo "Synapse did not start in time! Use \`$CONTAINER_RUNTIME logs synapse\` to investigate"
  exit 1
fi

set +e
REGISTER_USER_OUTPUT="$($CONTAINER_RUNTIME exec synapse register_new_matrix_user -u upschooling -p secret -a -c /data/homeserver.yaml http://localhost:8008)"
REGISTER_USER_SUCCESS=$?
set -e

if [[ "$REGISTER_USER_SUCCESS" != "0" ]]; then
  if echo $REGISTER_USER_OUTPUT | grep -q "User ID already taken."; then
    echo -e "User @upschooling:synapse already exists. \e[1;38;5;2mOK\033[0m"
  else
    echo "Could not create user @upschooling:synapse"
    echo $REGISTER_USER_OUTPUT
    exit 1
  fi
else
  echo -e "Matrix user @upschooling:synapse created. \e[1;38;5;2mOK\033[0m"
fi
